(*  title     : An encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

theory FOTheorems
 imports
   Theorems
begin
 sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="vampire cvc4 z3 e spass  " ]


section \<open>First Order ITL theorems\<close>

text {*

We give the proofs of a list of first order ITL theorems. 
*}


lemmas EExI_unl = EExI[unlift_rule] \<comment> \<open>@{text "w \<Turnstile> F x \<Longrightarrow> w \<Turnstile> (\<exists>\<exists> x. F x)"}\<close>


lemma EExNoDep:
 "\<turnstile> (\<exists>\<exists> x. G) = G"
proof -
  have 1: "\<turnstile> G \<longrightarrow> (\<exists>\<exists> x. G)" by (meson EExI)
  have 2: "\<And> x. \<turnstile> G \<longrightarrow> G" by simp
  have 3: "\<turnstile> (\<exists>\<exists> x. G) \<longrightarrow> G " using "2" by (meson EExE)
  from 1 3 show ?thesis using int_iffI by blast
qed

lemma AAxNoDep:
 "\<turnstile> (\<forall>\<forall> x. G) = G"
using EExNoDep 
by (metis (mono_tags, lifting) AAxDef EExE EExI int_iffI intensional_simps(4) inteq_reflection)


lemma EExEqvRule:
 assumes "\<And> x. \<turnstile> F x = G x"
 shows   "\<turnstile> (\<exists>\<exists> x. F x) = (\<exists>\<exists> x. G x)"
by (metis EExE EExI assms int_iffD1 int_iffD2 int_iffI lift_imp_trans)
 

lemma AAxImpEEx:
 "\<turnstile> (\<forall>\<forall> x. F x) \<longrightarrow> (\<exists>\<exists> x. F x)"
by (smt AAxDef EExI_unl Valid_def intensional_simps(4) inteq_reflection unl_lift unl_lift2)


lemma EExImpRule:
 assumes "\<turnstile> F x \<longrightarrow> G x"
 shows   "\<turnstile> (\<exists>\<exists> x. F x \<longrightarrow>  G x)  "
using assms by (meson MP EExI)

lemma EExImpRuleDist:
 assumes "\<turnstile> F x \<longrightarrow> G x"
 shows   "\<turnstile> (\<forall>\<forall> x. F x) \<longrightarrow> (\<exists>\<exists> x. G x) "
proof -
 have 1: "\<turnstile> (F x)  \<longrightarrow> (\<exists>\<exists> x. G x)" using EExI assms lift_imp_trans by blast 
 have 2: "\<turnstile>  \<not>(F x) \<or> (\<exists>\<exists> x. G x)" using "1" by auto
 have 3: "\<turnstile> \<not>(F x) \<longrightarrow> (\<exists>\<exists> x. \<not>(F x))" by (meson EExI)
 have 4: "\<turnstile> (\<exists>\<exists> x. \<not>(F x)) = \<not>(\<forall>\<forall> x. F x)" using AAxDef by fastforce
 from 2 3 4 show ?thesis by fastforce
qed

lemma EExImpNoDepDist:
 assumes "\<turnstile> R \<longrightarrow> G x"
 shows   "\<turnstile> R \<longrightarrow> (\<exists>\<exists> x. G x)"
using assms by (metis EExI lift_imp_trans)

lemma EExOrDist_1:
 "\<turnstile> (\<exists>\<exists> x. H x) \<longrightarrow> (\<exists>\<exists> x. (F x) \<or> (H x)) "
proof -
 have 1: "\<And> x. \<turnstile> H x \<longrightarrow> F x \<or> H x " by (simp add: Valid_def) 
 have 2: "\<And> x. \<turnstile> F x \<or> H x \<longrightarrow> (\<exists>\<exists> x. (F x) \<or> (H x))"  by (meson EExI)
 have 3: "\<And> x. \<turnstile> H x \<longrightarrow> (\<exists>\<exists> x. (F x) \<or> (H x))" using "1" "2" by (meson lift_imp_trans)
 from 3 show ?thesis using EExE by blast
qed

lemma EExOrDist_2:
 "\<turnstile> (\<exists>\<exists> x. F x) \<longrightarrow> (\<exists>\<exists> x. (F x) \<or> (H x)) "
proof -
 have 1: "\<And> x. \<turnstile> F x \<longrightarrow> F x \<or> H x " by (simp add: Valid_def) 
 have 2: "\<And> x. \<turnstile> F x \<or> H x \<longrightarrow> (\<exists>\<exists> x. (F x) \<or> (H x)) " by (meson EExI)
 have 3: "\<And> x. \<turnstile> F x \<longrightarrow> (\<exists>\<exists> x. (F x) \<or> (H x))" using "1" "2" by (meson lift_imp_trans)    
 from 3 show ?thesis using EExE by blast
qed

lemma EExOrDist_3:
 "\<turnstile> (\<exists>\<exists> x. F x) \<or> (\<exists>\<exists> x. H x) \<longrightarrow> (\<exists>\<exists> x. (F x) \<or> (H x)) "
using EExOrDist_2 EExOrDist_1 by fastforce

lemma EExOrDist_4:
 " \<turnstile> (\<exists>\<exists> x. (F x) \<or> (H x)) \<longrightarrow> (\<exists>\<exists> x. F x) \<or> (\<exists>\<exists> x. H x)"
proof -
 have 1: "\<And> x. \<turnstile> (F x) \<or> (H x) \<longrightarrow> (\<exists>\<exists> x. F x) \<or> (\<exists>\<exists> x. H x) " 
 by (metis EExI TrueW intensional_simps(11) intensional_simps(25) Prop02 Prop05 Prop08)
 from 1 show ?thesis by (simp add: EExE)
qed

lemma EExOrDist:
 "\<turnstile> ((\<exists>\<exists> x. F x) \<or> (\<exists>\<exists> x. H x)) = (\<exists>\<exists> x. (F x) \<or> (H x))"
using  EExOrDist_3 EExOrDist_4 by fastforce

lemma EExOrImport_1:
 "\<turnstile> G \<longrightarrow> (\<exists>\<exists> x. G \<or> (F x))"
by (simp add: EExI_unl Valid_def)

lemma EExOrImport_2:
 "\<turnstile> (\<exists>\<exists> x. F x) \<longrightarrow> (\<exists>\<exists> x. G \<or> (F x))"
by (simp add: EExOrDist_1)

lemma EExOrImport_3:
 "\<turnstile> (G \<or> (\<exists>\<exists> x. F x)) \<longrightarrow> (\<exists>\<exists> x. G \<or> (F x))"
using EExOrImport_1 EExOrImport_2  by fastforce

lemma EExOrImport_4:
 "\<turnstile> (\<exists>\<exists> x. G \<or> F x) \<longrightarrow> (G \<or> (\<exists>\<exists> x. F x))"
proof -
 have 1: "\<And> x. \<turnstile> G \<or> F x \<longrightarrow> G \<or> (\<exists>\<exists> x. F x)" by (meson EExI int_iffD2 int_simps(27) Prop04 Prop08)
 from 1 show ?thesis by (simp add: EExE)
qed

lemma EExOrImport:
 "\<turnstile> (G \<or> (\<exists>\<exists> x. F x)) = (\<exists>\<exists> x. G \<or> F x)"
by (simp add: EExOrImport_3 EExOrImport_4 int_iffI)


lemma EExAndImport_1:
 "\<turnstile> G \<and> (\<exists>\<exists> x. F x) \<longrightarrow> (\<exists>\<exists> x. G \<and> F x)"
proof -
 have 1: "\<turnstile> (G \<and> (\<exists>\<exists> x. F x) \<longrightarrow> (\<exists>\<exists> x. G \<and> F x))=
             ((\<exists>\<exists> x. F x) \<longrightarrow> (G \<longrightarrow> (\<exists>\<exists> x. G \<and> F x))) " by fastforce
 have 2: "\<And> x. \<turnstile> F x \<longrightarrow> (G \<longrightarrow> (\<exists>\<exists> x. G \<and> F x))" by (metis EExI int_eq lift_and_com Prop09)
 hence 3: "\<turnstile> (\<exists>\<exists> x. F x) \<longrightarrow> (G \<longrightarrow> (\<exists>\<exists> x. G \<and> F x))  " by (simp add: EExE)
 from 1 3 show ?thesis by auto
qed

lemma EExAndImport_2:
 "\<turnstile> (\<exists>\<exists> x. G \<and> F x) \<longrightarrow> G \<and> (\<exists>\<exists> x. F x)"
proof -
 have 1: "\<And> x. \<turnstile> G \<and> F x \<longrightarrow> G \<and> (\<exists>\<exists> x. F x) " 
 by (metis EExI int_iffD2 lift_and_com lift_imp_trans Prop12) 
 from 1 show ?thesis by (simp add: EExE)
qed

lemma EExAndImport:
 "\<turnstile> (G \<and> (\<exists>\<exists> x. F x)) = (\<exists>\<exists> x. G \<and> F x)"
by (simp add: EExAndImport_1 EExAndImport_2 int_iffI)


lemma EExAndDist:
 assumes "\<turnstile> F x \<and> G x"
 shows   "\<turnstile> (\<exists>\<exists> x. F x) \<and> (\<exists>\<exists> x. G x)"
proof -
 have 1: "\<turnstile> F x " using assms by fastforce
 have 2: "\<turnstile> G x " using assms by fastforce
 have 3: "\<turnstile> (\<exists>\<exists> x. F x)" using "1" by (meson EExI MP)  
 have 4: "\<turnstile> (\<exists>\<exists> x. G x)" using "2" by (meson EExI MP) 
 from 3 4 show ?thesis by fastforce
qed


lemma EExAndNoDepDist:
 assumes "\<turnstile> R \<and> G x"
 shows   "\<turnstile> R \<and> (\<exists>\<exists> x. G x)"
proof -
 have 1: "\<turnstile> R " using assms by fastforce
 have 2: "\<turnstile> G x" using assms by fastforce
 have 3: "\<turnstile> (\<exists>\<exists> x. G x)" using "2" by (meson EExI MP)
 from 1 3 show ?thesis by fastforce
qed

lemma Spec:
 "\<turnstile> (\<forall>\<forall> x. F x) \<longrightarrow> F x"
proof -
 have 1: "\<turnstile> \<not>(F x) \<longrightarrow> (\<exists>\<exists> x. \<not>(F x))" by (meson EExI)
 have 2: "\<turnstile> \<not>(\<exists>\<exists> x. \<not>(F x)) \<longrightarrow> F x" using 1 by auto
 from 2 show ?thesis using AAxDef by fastforce
qed

lemma AAxE:
 assumes "\<turnstile> (\<forall>\<forall> x. F x)"
         "\<turnstile> F x \<longrightarrow> R"
 shows   "\<turnstile> R"
using MP Spec assms(1) assms(2) by blast

lemma AAxI:
 assumes "\<And> x. \<turnstile> F x"
 shows   "\<turnstile> (\<forall>\<forall> x. F x)"
unfolding AAxDef 
by (smt AAxDef EExE assms intensional_simps(14) intensional_simps(4) inteq_reflection) 

lemma AAxEqvRule:
 assumes "\<And> x. \<turnstile> F x = G x"
 shows   "\<turnstile> (\<forall>\<forall> x. F x) = (\<forall>\<forall> x. G x)"
by (metis (mono_tags, lifting) AAxDef EExEqvRule assms int_iffD1 int_iffI 
    inteq_reflection lift_imp_neg)


lemma AAxAndDist:
 "\<turnstile> (\<forall>\<forall> x. (F x) \<and> (G x)) = ((\<forall>\<forall> x. F x) \<and> (\<forall>\<forall> x. G x))"
proof -
 have 1: "\<turnstile> ((\<exists>\<exists> x. \<not>(F x)) \<or> (\<exists>\<exists> x. \<not>(G x))) = (\<exists>\<exists> x. \<not>(F x) \<or> \<not>(G x))" by (simp add:EExOrDist)
 have 2: "\<turnstile> ((\<exists>\<exists> x. \<not>(F x))) = (\<not>(\<forall>\<forall> x. F x) ) " using AAxDef by fastforce  
 have 3: "\<turnstile> ((\<exists>\<exists> x. \<not>(G x))) = (\<not>(\<forall>\<forall> x. G x) ) " using AAxDef by fastforce
 have 4: "\<turnstile> ((\<exists>\<exists> x. \<not>(F x)) \<or> (\<exists>\<exists> x. \<not>(G x))) = (\<not>(\<forall>\<forall> x. F x) \<or> \<not>(\<forall>\<forall> x. G x) ) "
      using "2" "3" by fastforce
 have 5: "\<And> x . \<turnstile> (\<not>(F x) \<or> \<not>(G x)) = (\<not>((F x) \<and> (G x)))" by auto
 have 6: "\<turnstile> (\<exists>\<exists> x. \<not>(F x) \<or> \<not>(G x)) = (\<exists>\<exists> x. \<not>((F x) \<and> (G x))) " using "5" by (simp add: EExEqvRule)
 have 7: "\<turnstile> (\<exists>\<exists> x. \<not>((F x) \<and> (G x))) = (\<not>(\<forall>\<forall> x. (F x) \<and> (G x)))" using AAxDef by fastforce 
 have 8: "\<turnstile> (\<not>(\<forall>\<forall> x. F x) \<or> \<not>(\<forall>\<forall> x. G x) ) = \<not>( (\<forall>\<forall> x. F x) \<and> (\<forall>\<forall> x. G x))" by fastforce
 have 9: "\<turnstile> \<not>( (\<forall>\<forall> x. F x) \<and> (\<forall>\<forall> x. G x)) = (\<not>(\<forall>\<forall> x. (F x) \<and> (G x)))"
      using "1" "4" "6" "7" "8" by fastforce 
 from 9 show ?thesis by fastforce
qed

lemma AAxAndImport:
 "\<turnstile> (G \<and> (\<forall>\<forall> x. F x)) = (\<forall>\<forall> x. G \<and> F x)"
proof -
 have 1: "\<turnstile> (\<not> G \<or> (\<exists>\<exists> x. \<not>(F x))) = (\<exists>\<exists> x. \<not> G \<or> \<not>(F x)) " by (simp add: EExOrImport)
 have 2: "\<turnstile> ( (\<exists>\<exists> x. \<not>(F x))) = \<not>((\<forall>\<forall> x. F x))" using AAxDef by fastforce 
 have 3: "\<turnstile> ( \<not> G \<or> (\<exists>\<exists> x. \<not>(F x))) = \<not>(G \<and> (\<forall>\<forall> x. F x))" using "2" by fastforce
 have 4: "\<And> x. \<turnstile> (\<not> G \<or> \<not>(F x)) = \<not>(G \<and> F x)" by auto
 have 5: "\<turnstile> (\<exists>\<exists> x. \<not> G \<or> \<not>(F x)) = (\<exists>\<exists> x. \<not>(G \<and> F x))" using "4" by (simp add: EExEqvRule)
 have 6: "\<turnstile> (\<exists>\<exists> x. \<not>(G \<and> F x)) = \<not>(\<forall>\<forall> x. G \<and> F x)" using AAxDef by fastforce 
 have 7: "\<turnstile> \<not>(G \<and> (\<forall>\<forall> x. F x)) = \<not>(\<forall>\<forall> x. G \<and> F x)" using "1" "3" "5" "6" by fastforce
 from 7 show ?thesis by fastforce
qed

lemma AAxOrImport:
 "\<turnstile> (G \<or> (\<forall>\<forall> x. F x)) = (\<forall>\<forall> x. G \<or> F x)"
proof -
 have 1: "\<turnstile> (\<not> G \<and> (\<exists>\<exists> x. \<not>(F x))) = (\<exists>\<exists> x. \<not> G \<and> \<not>(F x))" by (simp add: EExAndImport)
 have 2: "\<turnstile> (\<exists>\<exists> x. \<not>(F x)) =  \<not>((\<forall>\<forall> x. F x))" using AAxDef by fastforce 
 have 3: "\<turnstile> ( \<not> G \<and> (\<exists>\<exists> x. \<not>(F x))) = \<not>(G \<or> (\<forall>\<forall> x. F x))" using "2" by fastforce
 have 4: "\<And> x. \<turnstile> (\<not> G \<and> \<not>(F x)) = \<not>(G \<or> F x)" by auto
 have 5: "\<turnstile> (\<exists>\<exists> x. \<not> G \<and> \<not>(F x)) = (\<exists>\<exists> x. \<not>(G \<or> F x))" using "4" by (simp add: EExEqvRule)
 have 6: "\<turnstile> (\<exists>\<exists> x. \<not>(G \<or> F x)) = \<not>(\<forall>\<forall> x. G \<or> F x)" using AAxDef by fastforce 
 have 7: "\<turnstile> \<not>(G \<or> (\<forall>\<forall> x. F x)) = \<not>(\<forall>\<forall> x. G \<or> F x)" using "1" "3" "5" "6" by fastforce
 from 7 show ?thesis by auto
qed

lemma EExImpChopRule:
 assumes "\<turnstile> F x \<longrightarrow> G x"
 shows   "\<turnstile> (\<exists>\<exists> x. H;(F x) \<longrightarrow> H;(G x))"
using RightChopImpChop EExImpRule assms by (smt MP EExI)

lemma EExChopRight:
 "\<turnstile> (\<exists>\<exists> x. (F x);F1) \<longrightarrow> (\<exists>\<exists> x. F x);F1"
proof -
 have 1: "\<And>x. \<turnstile> (F x);F1 \<longrightarrow> (\<exists>\<exists> x. F x);F1" by (simp add: EExI LeftChopImpChop)
 from 1 show ?thesis by (simp add: EExE)
qed

lemma EExChopRightNoDep:
 "\<turnstile> (\<exists>\<exists> x. (F x);F1) = (\<exists>\<exists> x. (F x));F1"
by (simp add: exist_state_d_def Valid_def chop_defs, auto)

lemma EExChopLeft : 
  "\<turnstile> (\<exists>\<exists> x. F1;(F x) ) \<longrightarrow> F1;(\<exists>\<exists> x. F x)"
proof -
 have 1: "\<And> x. \<turnstile> F1;(F x) \<longrightarrow> F1;(\<exists>\<exists> x. F x)" by (simp add: EExI RightChopImpChop)
 from 1 show ?thesis by (simp add: EExE)
qed

lemma EExChopLeftNoDep:
 "\<turnstile> (\<exists>\<exists> x. F1;(F x) ) = F1;(\<exists>\<exists> x. F x)"
by (simp add: exist_state_d_def Valid_def chop_defs, auto)

lemma  EExEExChopEqvEExEExChop:
 "\<turnstile> (\<exists>\<exists> v. (\<exists>\<exists> y. (F2 v);(F3 y) )) = (\<exists>\<exists> y. (\<exists>\<exists> v. (F2 v);(F3 y) )) "
by (simp add: exist_state_d_def Valid_def chop_defs, blast)

lemma  EExEExChopEqvEExChopEExA:
 "\<turnstile> (\<exists>\<exists> v. (\<exists>\<exists> y. (F2 v);(F3 y) )) = (\<exists>\<exists> v. (F2 v);(\<exists>\<exists> y. (F3 y) ) ) "
by (simp add: exist_state_d_def Valid_def chop_defs, blast)

lemma EExEExChopEqvEExChopEExB:
 "\<turnstile> (\<exists>\<exists> y. (\<exists>\<exists> v. (F2 v);(F3 y) )) = (\<exists>\<exists> y. (\<exists>\<exists> v. (F2 v)); (F3 y)) "
by (simp add: exist_state_d_def Valid_def chop_defs, blast)

lemma EExEExChopEqvEExChopEExC:
 "\<turnstile> (\<exists>\<exists> v. (\<exists>\<exists> y. (F2 v);(F3 y) )) = (\<exists>\<exists> v. (F2 v));(\<exists>\<exists> y. (F3 y))"
by (metis EExChopRightNoDep EExEExChopEqvEExChopEExA EExNoDep Prop04)

lemma AAxRev:
 "\<turnstile> (\<forall>\<forall> x. F x)\<^sup>r = (\<forall>\<forall> x. (F x)\<^sup>r)"
proof -
 have 1: "\<turnstile> (\<forall>\<forall> x. F x) = (\<not>(\<exists>\<exists> x. \<not>(F x)))" using AAxDef by blast
 have 2: "\<turnstile> (\<forall>\<forall> x. F x)\<^sup>r = (\<not>(\<exists>\<exists> x. \<not>(F x)))\<^sup>r" using REqvRule "1" by blast
 have 3: "\<turnstile> (\<not>(\<exists>\<exists> x. \<not>(F x)))\<^sup>r = \<not>((\<exists>\<exists> x.(\<not> (F x)))\<^sup>r)" by (simp add: rev_fun1)
 have 4: "\<turnstile> ((\<exists>\<exists> x.(\<not> (F x)))\<^sup>r) = ((\<exists>\<exists> x.(\<not> (F x))\<^sup>r))" by (simp add: EExRev)
 hence 5: "\<turnstile> \<not>((\<exists>\<exists> x.(\<not> (F x)))\<^sup>r) = \<not>(\<exists>\<exists> x.(\<not> (F x))\<^sup>r) " by auto
 have 51: "\<And> x. \<turnstile> (\<not> (F x))\<^sup>r = \<not>( (F x)\<^sup>r)" by (simp add: rev_fun1)
 hence 52: "\<turnstile> (\<exists>\<exists> x.(\<not> (F x))\<^sup>r) = (\<exists>\<exists> x.\<not>( (F x)\<^sup>r))" using EExEqvRule by fastforce
 hence 6: "\<turnstile> \<not>(\<exists>\<exists> x.(\<not> (F x))\<^sup>r) = \<not>(\<exists>\<exists> x.\<not>( (F x)\<^sup>r))"   by fastforce
 have 7: "\<turnstile> \<not>(\<exists>\<exists> x.\<not>( (F x)\<^sup>r)) = (\<forall>\<forall> x. (F x)\<^sup>r) " using AAxDef by fastforce
 from 1 2 3 5 6 7 show ?thesis by fastforce
qed
 
lemma ExLen:
 "\<turnstile> \<exists> n. len(n)"
by (simp add: Valid_def len_defs)

lemma CSPowerChop:
 "\<turnstile> (f\<^sup>\<star>) = (\<exists>n. powerchop f n)"
by (simp add: chopstar_eqv_power_chop Valid_def) 

lemma ExChopRightNoDep:
 "\<turnstile> (\<exists> x. (F x);F1) = (\<exists> x. (F x));F1"
by (simp add: Valid_def chop_defs, auto)

lemma ExChopLeftNoDep:
 "\<turnstile> (\<exists> x. F1;(F x) ) = F1;(\<exists> x. F x)"
by (simp add: Valid_def chop_defs, auto)

lemma ExExEqvExEx:
 "\<turnstile> (\<exists> x. (\<exists> y. (F1 x);(F2 y))) = (\<exists> y. (\<exists> x. (F1 x);(F2 y)))"
by (simp add: Valid_def chop_defs, auto)



end
